#include "scene.h"

#include "V3.h"
#include "M33.h"
#include "tm.h"

Scene *scene;

using namespace std;

#include <iostream>
#include <fstream>

Scene::Scene() {


	gui = new GUI();
	gui->show();

	cgi = 0;
	soi = 0;

	int u0 = 16;
	int v0 = 40;
	int h = 400;
	int w = 600;
	fb = new FrameBuffer(u0, v0, w, h);
	fb->position(u0, v0);
	fb->label("first person");
	fb->show();
	fb->SetBGR(0xFF0000FF);
	fb->redraw();

	fb3 = new FrameBuffer(u0, v0, w, h);
	fb3->position(u0+fb->w+u0, v0);
	fb3->label("3rd person");
	fb3->show();
	fb3->SetBGR(0xFFFF0000);
	fb3->redraw();

	hwClearColor = V3(0.0f, 0.0f, 0.5f);
	hwfb = new FrameBuffer(u0, v0, w, h);
	hwfb->position(u0 + fb->w + u0, v0);
	hwfb->label("first person HW");
	hwfb->ishw = 1;
	hwfb->show();
	hwfb->redraw();
	fb3->hide();

	float hfov = 40.0f;
	ppc = new PPC(hfov, w, h);
	ppc3 = new PPC(hfov, w, h);
	ppc3->C = V3(-200.0f, 200.0f, 120.0f);

	tmsN = 2;
	tms = new TM[tmsN];
	tms[0].LoadBin("geometry/teapot1k.bin");
	tms[1].LoadBin("geometry/teapot1k.bin");
	//	tms[0].LoadBin("geometry/teapot57k.bin");
	tms[0].SetCenter(V3(0.0f, 0.0f, -150.0f));
	tms[1].SetTexturedRectangle(150.0f, 150.0f);
	tms[1].Rotate(tms[1].GetCenter(), V3(1.0f, 0.0, 0.0f), 90.0f);
	tms[1].SetCenter(tms[0].GetCenter());
	tms[1].Translate(V3(0.0f, -23.0f, 30.0f));
	FrameBuffer *text = new FrameBuffer(0, 0, 256, 256);
	text->SetChecker(0xFF000000, 0xFFFFFFFF, 16);
	tms[1].texture = text;

	ppc3->PositionAndOrient(ppc3->C, tms[0].GetCenter(), V3(0.0f, 1.0f, 0.0f));

	gui->uiw->position(u0, v0 + fb->h + v0);

	// load textures from image files
	textures = 0;
	texturesN = 0;

	needInitHW = 1;
	bwFraction = 0.0f;

}

void Scene::Render(PPC *rppc, FrameBuffer *rfb) {

	rfb->SetBGR(0xFFFFFFFF);
	rfb->SetZB(0.0f);
	for (int tmi = 0; tmi < tmsN; tmi++) {
//		tms[tmi].RenderWireFrame(rppc, rfb);
		tms[tmi].RenderFilled(rppc, rfb);
	}
	rfb->redraw();

}

void Scene::Render() {

	// first person
	Render(ppc, fb);
	// third person
	Render(ppc3, fb3);
	fb3->Render3DSegment(ppc3, ppc->C, ppc->C + ppc->GetVD()*100.0f, 
		V3(1.0f, 0.0f, 0.0f), V3(0.0f, 0.0f, 0.0f));
	fb3->Render3DPoint(ppc3, ppc->C, V3(1.0f, 0.0f, 0.0f), 7.0f);
	fb3->redraw();

}

#if 0
I = P + rt // t is a float (scalar) parameter
I = V0 + (V1 - V0)b + (V2 - V0)c; I = V0a + V1b + V2c
P + rt = V0 + (V1 - V0)b + (V2 - V0)c
P - V0 = (V1 - V0)b + (V2 - V0)c - rt
P - V0 = [(V1 - V0)   (V2 - V0) -r] x[b c t]T
P - V0 = M x [b c t]T
[b c t]T = M-1 x (P - V0)
If t < 0, no intersection
	If(b < 0 || c < 0 || (1 - b - c) < 0), no intersection
	I = P + rt;
// barycentric coordinates (a = 1-b-c, b, c)
// I = aV0 + bV1 + cV2
#endif

int Scene::IntersectTriangleWithRay(V3 v0, V3 v1, V3 v2, V3 rayO, V3 rayDir, V3 &abc, float &currt) {

	M33 M;
	M.SetColumn(0, v1 - v0);
	M.SetColumn(1, v2 - v0);
	M.SetColumn(2, rayDir*-1.0f);
	V3 bct = M.Inverted() * (rayO - v0);
	if (bct[2] < 0.001f)
		return 0;
	abc[0] = 1.0f - bct[0] - bct[1];
	abc[1] = bct[0];
	abc[2] = bct[1];
	if (abc[0] < 0.0f || abc[1] < 0.0f || abc[2] < 0.0f)
		return 0;
	currt = bct[2];
	return 1;

}

int Scene::RayTrace(V3 rayD, V3 rayO, V3 &colv, V3 &rrD, V3 &rrO) {

	int ret = 1;
	colv = V3(1.0f, 1.0f, 1.0f);
	float mint = FLT_MAX;
	for (int ti = 0; ti < tmsN; ti++) {
		for (int tri = 0; tri < tms[ti].trisN; tri++) {
			V3 v0 = tms[ti].verts[tms[ti].tris[3 * tri + 0]];
			V3 v1 = tms[ti].verts[tms[ti].tris[3 * tri + 1]];
			V3 v2 = tms[ti].verts[tms[ti].tris[3 * tri + 2]];
			V3 abc;
			float currt;
			if (IntersectTriangleWithRay(v0, v1, v2, rayO, rayD, abc, currt)) {
				if (currt > mint) {
					continue;
				}
				ret = 0;
				mint = currt;
				M33 cm;
				cm.SetColumn(0, tms[ti].cols[tms[ti].tris[3 * tri + 0]]);
				cm.SetColumn(1, tms[ti].cols[tms[ti].tris[3 * tri + 1]]);
				cm.SetColumn(2, tms[ti].cols[tms[ti].tris[3 * tri + 2]]);
				colv = cm*abc;
				M33 nm;
				nm.SetColumn(0, tms[ti].normals[tms[ti].tris[3 * tri + 0]]);
				nm.SetColumn(1, tms[ti].normals[tms[ti].tris[3 * tri + 1]]);
				nm.SetColumn(2, tms[ti].normals[tms[ti].tris[3 * tri + 2]]);
				V3 currn = nm*abc; currn = currn.UnitVector();

				if (tms[ti].texture) {
					M33 tcm;
					tcm.SetColumn(0, tms[ti].tcs[tms[ti].tris[3 * tri + 0]]);
					tcm.SetColumn(1, tms[ti].tcs[tms[ti].tris[3 * tri + 1]]);
					tcm.SetColumn(2, tms[ti].tcs[tms[ti].tris[3 * tri + 2]]);
					V3 tc = tcm*abc;
					colv.SetFromColor(tms[ti].texture->LookUpBilinear(tc[0], tc[1]));
				}

				rrD = currn.Reflect(rayD);
				rrO = rayO + rayD * currt;
			}
		}
	}

	return ret;

}

void Scene::RayTrace(PPC *ppc, FrameBuffer *fb) {
	for (int v = 0; v < fb->h; v++) {
		for (int u = 0; u < fb->w; u++)
			fb->Set(u, v, 0xFF0000FF);
		fb->redraw();
		Fl::check();
		for (int u = 0; u < fb->w; u++) {
			V3 rayD = (ppc->c + ppc->a*(.5f + (float)u) + ppc->b*(.5f + (float)v)).UnitVector();
			V3 rayO = ppc->C;
			V3 colv, rrD, rrO;
			int bgr = RayTrace(rayD, rayO, colv, rrD, rrO);
			V3 blendedColorv;
			if (bgr) {
				blendedColorv = colv;
			}
			else {
				V3 rrrD, rrrO, rcolv;
				bgr = RayTrace(rrD, rrO, rcolv, rrrD, rrrO);
				if (bgr) {
					blendedColorv = colv;
				}
				else {
					blendedColorv = (colv + rcolv) * 0.5f;
				}
			}

			fb->Set(u, v, blendedColorv.GetColor());
		}
	}
	fb->redraw();
}

float Randm11() {
	double tmp = rand();
	tmp /= RAND_MAX;
	tmp = tmp*2.0f - 1.0f;
	return (float)tmp;
}

void Scene::DBG() {
	{
		fb->LoadTiff("mydbg/table.tif");
		FrameBuffer *fb0 = new FrameBuffer(0, 0, 10, 10);
		fb0->LoadTiff("mydbg/table.tif");

		fb->redraw();
		hwfb->hide();
		fb3->hide();
		tmsN = 1;
		tms[0].SetTexturedRectangle(54.0f, 24.0f);
		tms[0].Rotate(V3(0.0f, 0.0f, 0.0f), V3(1.0f, 0.0f, 0.0f), 90.0f);
		float fov0 = 55.0f;
		ppc = new PPC(fov0, fb->w, fb->h);
		ppc->PositionAndOrient(V3(54.0f, 70.0f, -20.0f), V3(0.0f, 0.0f, 0.0f), V3(0.0f, 1.0f, 0.0f));

		// initial guess;
		PPC ppc0(*ppc);

		V3 trueProjections[4];
		trueProjections[0] = V3(134.0f, 89.0f, 0.0f);
		trueProjections[1] = V3(273.0f, 98.0f, 0.0f);
		trueProjections[2] = V3(271.0f, 324.0f, 0.0f);
		trueProjections[3] = V3(31.0f, 295.0f, 0.0f);

		float minError = FLT_MAX;
		// search space 3 translations and 3 rotations and field of view
		V3 transRange(40.0f, 40.0f, 40.0f);
		V3 rotRange(30.0f, 30.0f, 30.0f);
		float fovRange = 30.0f;
		int triesN = 1000000000;
		int tri = 0;
		while (tri < triesN) {

			// compute error
			float err = 0.0f;
			for (int vi = 0; vi < tms[0].vertsN; vi++) {
				V3 actualProjection;
				if (!ppc->Project(tms[0].verts[vi], actualProjection))
					err += 10000.0f;
				actualProjection[2] = 0.0f;
				err += (trueProjections[vi] - actualProjection).Length();
			}
			err /= 4.0f;
			if (err < minError) {


				// visualize guess
				fb->SetZB(0.0f);
				for (int uv = 0; uv < fb->w*fb->h; uv++)
					fb->pix[uv] = fb0->pix[uv];
				tms[0].RenderWireFrame(ppc, fb);
				tms[0].RenderAsPoints(7, ppc, fb);
				unsigned int cols[4] = { 0xFF0000FF, 0xFF00FF00, 0xFFFF0000, 0xFF00FFFF };
				for (int vi = 0; vi < 4; vi++)
					fb->DrawDisk(trueProjections[vi], 5, cols[vi]);
				//		tms[0].RenderFilled(rppc, rfb);

				cerr << "INFO: error: " << err << endl;
				fb->redraw();
				Fl::check();
				minError = err;
//				ppc0 = *ppc;
			}

			V3 transGuess(Randm11()*transRange[0], Randm11()*transRange[1], Randm11()*transRange[2]);
			V3 rotGuess(Randm11()*rotRange[0], Randm11()*rotRange[1], Randm11()*rotRange[2]);
			float fovGuess = Randm11()*fovRange;
			ppc = new PPC(fov0+fovGuess, ppc->w, ppc->h);
			*ppc = ppc0;
			ppc->C = ppc0.C + transGuess;
			ppc->Rotate(ppc->b*-1.0f, rotGuess[0]);
			ppc->Rotate(ppc->a, rotGuess[1]);
			ppc->Rotate(ppc->GetVD(), rotGuess[2]);
			tri++;
			if ((tri % 100000) == 0)
				cerr << tri << "           \r";
		}

		return;
	}
	{
		fb->SetBGR(0xFFFFFFFF);
		fb->redraw();
		Fl::check();
		hwfb->hide();
		RayTrace(ppc, fb);
		return;
	}
	{
		hwfb->hide();
		fb3->show();
		*ppc3 = *ppc;
		PPC *ppc0 = new PPC(*ppc);
		Render();
		fb->SaveAsTiff("mydbg/fb0.tif");
		FrameBuffer *fb0;
		fb0 = new FrameBuffer(0, 0, 10, 10);
		fb0->LoadTiff("mydbg/fb0.tif");
		int fN = 100;
		for (int fi = 0; fi < fN; fi++) {
			Render(ppc, fb);
			fb3->Rotate(fb0, ppc0, ppc3);
			fb3->redraw();
			ppc3->Rotate(ppc3->b*-1.0f, .1f);
//			ppc3->Zoom(1.01f);
			ppc3->Zoom(0.99f);
			*ppc = *ppc3;
			Fl::check();
		}
		return;
	}
	{
		AABB aabb(tms[0].SetAABB());
		cerr << aabb.minv << endl;
		cerr << aabb.maxv << endl;
		return;
	}

	{
		int fN = 100;
		for (int fi = 0; fi < fN; fi++) {
			Render();
			hwfb->redraw();
			Fl::check();
			bwFraction = (float)fi / (float)(fN - 1);
//			ppc->C = ppc->C + V3(0.0f, 0.1f, 0.0f);
//			ppc->PositionAndOrient(ppc->C, tms[0].GetCenter(), V3(0.0f, 1.0f, 0.0f));
		}
		return;

	}

	{
		hwfb->hide();
		fb3->show();
		FrameBuffer *texture = new FrameBuffer(100, 100, 1024, 1024);
		texture->SetChecker(0xFF000000, 0xFFDDDDDD, 2);
		ppc->PositionAndOrient(V3(0.0f, 0.0f, 0.0f), V3(0.0f, 0.0f, -100.0f), V3(0.0f, 1.0f, 0.0f));
		float rw = 200.0f;
		float rh = 200.0f;
		tms[0].SetTexturedRectangle(rw, rh);
		tms[0].Translate(V3(0.0f, 0.0f, -200.0f));
//		tms[0].Rotate(tms[0].GetCenter(), V3(0.0f, 0.0f, 1.0f), 5.0f);
		tms[0].texture = texture;
		//		Render();
		//		fb->SaveAsTiff("mydbg/bilinear.tif");
		//		return;

		for (int fi = 0; fi < 100000; fi++) {
			Render();
			Fl::check();
//			tms[0].Rotate(tms[0].GetCenter(), V3(0.0f, 1.0f, 0.0f), .1f);
			tms[0].Translate(V3(0.0f, 0.0f, -.1f));
		}
		Render();
		return;
	}


	{
		int fN = 10000;
		for (int fi = 0; fi < fN; fi++) {
			hwClearColor = V3(0.0f, 0.0f, (float)fi / (float)(fN - 1));
			hwfb->redraw();
			Fl::check();
		}
		return;
	}

	{
		Render();
		return;
	}


	{
		V3 v0(-100.0f, 50.0f, -200.0f);
		V3 v1(100.0f, -50.0f, -200.0f);
		V3 v2(100.0f, 50.0f, -200.0f);
		tms[0].SetOneTriangle(v0, v1, v2);
		ppc->PositionAndOrient(V3(0.0f, 0.0f, 0.0f), V3(0.0f, 0.0f, -200.0f),
			V3(0.0f, 1.0f, 0.0f));
		for (int fi = 0; fi < 360; fi++) {
			tms[0].Rotate(tms[0].GetCenter(), V3(0.0f, 0.0f, -1.0f), 1.0f);
			Render();
			tms[0].RenderWireFrame(ppc, fb);
			Fl::check();
		}
		return;
	}

	{
		TM tm;
		tm.SetAsCopy(&(tms[0]));
		tms[0].SetUnshared(&tm);
		for (int fi = 0; fi < 1000; fi++) {
			Render();
			Fl::check();
			tms[0].Explode(0.1f);
		}
		return;
	}

	{
		Render();
		float vlength = 10.0f;
		tms[0].VisualizeVertexNormals(ppc, fb, vlength);
		tms[0].VisualizeVertexNormals(ppc3, fb3, vlength);
		return;

	}

	{
		V3 matColor(1.0f, 0.0f, 0.0f);
		float ka = 0.03f;
		V3 lv = ppc->GetVD();
		V3 center = tms[0].GetCenter();
		for (int fi = 0; fi < 360; fi++) {
			tms[0].Light(matColor, ka, lv);
			Render();
			fb3->Render3DSegment(ppc, center - lv *50.0f, center, V3(1.0f, 1.0f, 0.0f),
				V3(1.0f, 0.0f, 0.0f));
			Fl::check();
			lv = lv.RotateThisVector(V3(1.0f, 0.0f, 0.0f), 1.0f);

		}
		return;
	}

	{

		V3 P0(-5.0f, 0.0f, -10.0f);
		V3 P1(+100.0f, 0.0f, -200.0f);
		V3 c0(1.0f, 0.0f, 0.0f);
		V3 c1(0.0f, 0.0f, 0.0f);
		ppc->PositionAndOrient(V3(0.0f, 0.0f, 0.0f), V3(0.0f, 0.0f, -10.0f), V3(0.0f, 1.0f, 0.0f));
		fb->SetBGR(0xFFFFFFFF);
		fb->Render3DSegment(ppc, P0, P1, c0, c1);
		fb->redraw();
		ppc3->PositionAndOrient(V3(0.0f, 500.0f, 100.0f), V3(0.0f, 0.0f, 0.0f), V3(0.0f, 1.0f, 0.0f));
		fb3->SetBGR(0xFFFFFFFF);
		fb3->Render3DSegment(ppc3, P0, P1, c0, c1);
		int stepsN = 4;
		for (int si = 0; si <= stepsN; si++) {
			V3 currp = P0 + (P1 - P0) * (float) si / (float)stepsN;
			V3 currc = c0 + (c1 - c0) * (float) si / (float)stepsN;
			fb->Render3DPoint(ppc, currp, currc, 9.0f);
			fb3->Render3DPoint(ppc3, currp, currc, 9.0f);
		}
		V3 P0p, P1p;
		ppc->Project(P0, P0p);
		ppc->Project(P1, P1p);
		float w0 = 1.0f / P0p[2];
		float w1 = 1.0f / P1p[2];
		for (int si = 0; si <= stepsN; si++) {
			float frac = (float)si / (float)stepsN;
			V3 currp = P0p + (P1p - P0p) * frac;
			float currw = w0 + (w1 - w0)*frac;
			fb->DrawDisk(currp, 5.0f, 0xFF00FF00);
			V3 ucurrp;
			ppc->Unproject(currp, ucurrp);
			fb3->Render3DPoint(ppc3, ucurrp, V3(0.0f, 1.0f, 0.0f), 5.0f);
			currp[2] = 1.0f / currw;
			ppc->Unproject(currp, ucurrp);
			fb3->Render3DPoint(ppc3, ucurrp, V3(0.0f, 0.0f, 1.0f), 5.0f);
		}
		fb->redraw();
		fb3->redraw();
		return;
	}

	{
		for (int fi = 0; fi < 100; fi++) {
			Render();
			Fl::check();
			ppc->C = ppc->C + V3(0.0f, 0.5f, 0.0f);
			ppc->PositionAndOrient(ppc->C, tms[0].GetCenter(), V3(0.0f, 1.0f, 0.0f));
		}
		return;
	}

	{
		V3 cc(0.0f, 0.0f, -100.0f);
		float cubeSideLength = 30.0f;
		TM tm;
//		tm.SetAsAACube(cc, cubeSideLength);
		tm.LoadBin("geometry/teapot57k.bin");
		tm.SetCenter(V3(0.0f, 0.0f, -150.0f));
		int psize = 5;
		V3 Cn(ppc->C);
		V3 L = tm.GetCenter();
		V3 upg(0.0f, 1.0f, 0.0f);
		PPC ppc0(*ppc);
		int fN = 36;
//		fN = 1;
		for (int fi = 0; fi < fN; fi++) {
			fb->SetBGR(0xFFFFFFFF);
			fb->SetZB(0.0f);
//			tm.RenderAsPoints(psize, ppc, fb);
			tm.RenderWireFrame(ppc, fb);
			fb->redraw();
			Fl::check();
			float tstep = 0.1f;
			V3 tv = ppc->GetVD()*tstep;
			tv = ppc->a*tstep;
			V3 center = tm.GetCenter();
			ppc->Rotate(ppc->b.UnitVector()*-1.0f, 0.01f);
//			tm.Rotate(center, V3(0.0f, 1.0f, 0.0f), 0.1f);
//			ppc->C = ppc->C + tv;
			ppc->PositionAndOrient(Cn, L, upg);
			Cn = Cn + V3(0.0f, 5.0f, 0.0f);			
		}
		*ppc = ppc0;
		return;
	}
	{
		V3 P(0.0f, 0.0f, -100.0f);
		V3 pP;
		ppc->Project(P, pP);
		cerr << pP << endl;
		return;
	}

	{

		fb->SetBGR(0xFF000000);
		fb->SaveAsTiff("mydbg/blackImage.tif");
		fb->SetBGR(0xFF00FF00);
		fb->redraw();
		Fl::check();
		char c;
		cin >> c;
		fb->LoadTiff("mydbg/redImage.tif");
		fb->redraw();
		return;
	}

	{

		V3 p0(20.0f, 430.0f, 0.0f);
		V3 p1(520.0f, 331.0f, 0.0f);
		fb->SetBGR(0xFFFFFFFF);
		fb->DrawSegment(p0, p1, 0xFF000000);
		fb->redraw();
		return;

	}

	{
		V3 tlc(100.5f, 45.5f, 0.0f);
		V3 brc(300.5f, 145.5f, 0.0f);
		unsigned int col = 0xFF0000FF;
		unsigned int colDisk = 0xFFFFFF00;
		V3 center(114.1f, 300.1f, 0.0f);
		float r = 110.0f;
		int fN = 1000;
		float du = 1.f;
		for (int fi = 0; fi < fN; fi++) {
			fb->SetBGR(0xFFFFFFFF);
			fb->DrawAARectangle(tlc, brc, col);
//			fb->DrawDisk(center, r, colDisk);
			fb->redraw();
			Fl::check();
			tlc[0] += du;
			brc[0] += du;
			center[0] += du;
		}
		return;
	}

	{
		M33 m;
		V3 v(1.0f, 2.0f, 3.0f);
		m[0] = v;
		m[1] = v;
		m[2] = v;
		V3 ret = m*v;
		cerr << "mv: " << ret << endl;
		return;
	}
	{
		V3 v(1.0f, 2.0f, 3.0f);
		cerr << "v: " << v << endl;
		return;
	}

	cerr << "INFO: pressed DBG button on GUI" << endl;

	{
		V3 v0(1.0f, 2.0f, 3.0f);
		V3 v1(-1.0f, -2.0f, -3.0f);
		V3 v01 = v0 + v1;
		cerr << "v0 + v1: " << v01[0] << " " << v01[1] << " " << v01[2] << endl;
		return;
	}

	return;

	unsigned int col0 = 0xFF000000;
	unsigned int col1 = 0xFFFFFFFF;
	int csize = 100;
	fb->SetChecker(col0, col1, csize);
	fb->redraw();
}

void Scene::NewButton() {
	cerr << "INFO: pressed New button on GUI" << endl;
}


void Scene::RenderHW() {

	int shaders = 1;
	if (needInitHW) {
		glEnable(GL_DEPTH_TEST);
		// create HW textures
		// create texture handles
		// set parameters
		// pass textures[ti].pix pixels to HW to know the texels of each texture
		if (shaders) {
			cgi = new CGInterface();
			cgi->PerSessionInit();
			soi = new ShaderOneInterface();
			soi->PerSessionInit(cgi);
		}
		needInitHW = 0;
	}

	if (shaders) {
		cgi->EnableProfiles();
		soi->PerFrameInit();
	}

	glClearColor(hwClearColor[0], hwClearColor[1], hwClearColor[2], 1.0f);
	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	float nearz = 0.1f;
	float farz = 1000.0f;
	ppc->SetIntrinsicsHW(nearz, farz);
	ppc->SetExtrinsicsHW();
	for (int tmi = 0; tmi < tmsN; tmi++) {
		tms[tmi].RenderHW();
	}

	if (shaders)
		cgi->DisableProfiles();

}